@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2008-2022 Sergey Poznyakoff
+   Copyright (C) 2008-2025 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,7 +36,8 @@ enum pies_command
     COM_STOP,
     COM_DUMP_DEPMAP,
     COM_TRACE_PREREQ,
-    COM_TRACE_DEPEND
+    COM_TRACE_DEPEND,
+    COM_LIST_SHUTDOWN_SEQUENCE
   };
 
 enum pies_command command;
@@ -323,15 +324,15 @@ static struct tokendef sig_tokendef[] = {
 void
 action_free (struct action *act)
 {
-  if (!act)
-    return;
-  if (act->nstat > 0)
-    free (act->status);
-  free (act->addr);
-  free (act->message);
-  free (act->command);
+  if (act)
+    {
+      grecs_list_free (act->cond_list);
+      free (act->addr);
+      free (act->message);
+      free (act->command);
 
-  free (act);
+      free (act);
+    }
 }
 
 static void
@@ -340,6 +341,35 @@ free_entry_action (void *act)
   action_free (act);
 }
 
+int
+strtosignum (char const *arg)
+{
+  size_t len;
+  int n;
+
+  if (arg == NULL)
+    return 0;
+
+  len = strlen (arg);
+  if (len > 3 && memcmp (arg, "SIG", 3) == 0)
+    {
+      if (arg[4] == '+')
+	{
+	  char *p;
+	  unsigned long un;
+
+	  errno = 0;
+	  un = strtoul (arg + 4, &p, 0);
+	  if (un == 0 || *p || errno || un > INT_MAX)
+	    return 0;
+	  n = un;
+	}
+      else if (strtotok_ci (sig_tokendef, arg, &n))
+	return 0;
+    }
+  return n;
+}
+
 static struct action *
 create_action (struct component *comp,
 	       grecs_locus_t *locus,
@@ -347,22 +377,38 @@ create_action (struct component *comp,
 	       const char *(*getarg) (grecs_value_t *, int, grecs_locus_t *))
 {
   int i;
-  unsigned *retv;
-  int retc = 0;
-  int allflag = 0;
+  struct grecs_list *cond_list;
   struct action *act;
 
-  retv = grecs_calloc (argc, sizeof *retv);
   if (argc == 0 || (argc == 1 && strcmp (getarg (val, 0, locus), "*") == 0))
-    allflag = 1;
+    cond_list = NULL;
   else
     {
+      cond_list = grecs_list_create ();
+      cond_list->free_entry = listel_dispose;
+      
       for (i = 0; i < argc; i++)
 	{
-	  unsigned n;
+	  struct status_cond *sc;
+	  int n;
+	  int neg = 0;
 	  const char *arg = getarg (val, i, locus);
 	  size_t len = strlen (arg);
 
+	  if (*arg == '!')
+	    {
+	      neg = 1;
+	      arg++;
+	      if (*arg == 0)
+		{
+		  i++;
+		  if (i == argc)
+		    break;
+		  arg = getarg (val, i, locus);
+		  len = strlen (arg);
+		}
+	    }
+	  
 	  if (isdigit (arg[0]))
 	    {
 	      char *p;
@@ -375,22 +421,14 @@ create_action (struct component *comp,
 	    }
 	  else if (len > 3 && memcmp (arg, "SIG", 3) == 0)
 	    {
-	      if (arg[4] == '+')
+	      int signo;
+	      if ((signo = strtosignum (arg)) == 0)
 		{
-		  char *p;
-		  n = strtoul (arg + 4, &p, 0);
-		  if (*p)
-		    {
-		      grecs_error (locus, 0, _("%s: not a number"), p);
-		      continue;
-		    }
-		}
-	      else if (strtotok_ci (sig_tokendef, arg, (int*) &n))
-		{
-		  grecs_error (locus, 0, _("%s: not a signal code"), arg);
+		  grecs_error (locus, 0, _("%s: bad signal specification"),
+			       arg);
 		  continue;
 		}
-	      n |= STATUS_SIG_BIT;
+	      n = STATUS_SIG_BIT | signo;
 	    }
 	  else if (strtotok_ci (ex_tokendef, arg, (int *) &n))
 	    {
@@ -398,23 +436,16 @@ create_action (struct component *comp,
 	      continue;
 	    }
 
-	  /* Alles in ordnung */
-	  retv[retc++] = n;
+	  sc = grecs_malloc (sizeof (sc[0]));
+	  sc->status = n;
+	  sc->neg = neg;
+
+	  grecs_list_append (cond_list, sc);
 	}
     }
 
-  if (retc == 0 && !allflag)
-    {
-      free (retv);
-      return NULL;
-    }
-
   act = grecs_zalloc (sizeof *act);
-  if (!allflag)
-    {
-      act->nstat = retc;
-      act->status = retv;
-    }
+  act->cond_list = cond_list;
   if (!comp->act_list)
     {
       comp->act_list = grecs_list_create ();
@@ -1246,6 +1277,27 @@ _cb_flags (enum grecs_callback_command cmd,
   return 0;
 }
 
+static int
+_cb_signal (enum grecs_callback_command cmd,
+	    grecs_node_t *node,
+	    void *varptr, void *cb_data)
+{
+  grecs_locus_t *locus = &node->locus;
+  grecs_value_t *value = node->v.value;
+  int n;
+
+  if (grecs_assert_node_value_type (cmd, node, GRECS_TYPE_STRING))
+    return 1;
+  if ((n = strtosignum (value->v.string)) == 0)
+    {
+      grecs_error (locus, 0, _("%s: bad signal specification"),
+		   value->v.string);
+      return 1;
+    }
+  *(int*)varptr = n;
+  return 0;
+}
+  
 struct grecs_keyword component_keywords[] = {
   {"mode",
    N_("mode"),
@@ -1285,6 +1337,11 @@ struct grecs_keyword component_keywords[] = {
    N_("List of flags."),
    grecs_type_string, GRECS_LIST, NULL, offsetof (struct component, flags),
    _cb_flags },
+  {"sigterm",
+   N_("SIGNAL"),
+   N_("Termination signal."),
+   grecs_type_string, GRECS_DFLT, NULL, offsetof (struct component, sigterm),
+   _cb_signal },
 #if PIES_SYSVINIT_ENABLED
   {"runlevels",
    N_("chars"),
@@ -1584,6 +1641,9 @@ static struct grecs_keyword syslog_kw[] = {
   {"tag", N_("string"), N_("Tag syslog messages with this string"),
    grecs_type_string, GRECS_DFLT,
    &pies_log_tag},
+  {"fallback-file", N_("filename"), N_("Name of the fallback log file"),
+   grecs_type_string, GRECS_DFLT,
+   &pies_fallback_file},
 #if 0
   /* This is reserved for future use */
   {
@@ -1896,7 +1956,15 @@ pies_schedule_children (int op)
   children_op |= op;
 }
 
-RETSIGTYPE
+static int
+children_op_toggle (int op)
+{
+  int result = children_op & op;
+  children_op &= ~op;
+  return result;
+}
+
+void
 sig_handler (int sig)
 {
   if (SYSVINIT_ACTIVE && sysvinit_sigtrans (sig, &action))
@@ -1930,7 +1998,7 @@ sig_handler (int sig)
     }
 }
 
-RETSIGTYPE
+void
 sigchld_early (int sig)
 {
   while (waitpid (-1, NULL, WNOHANG) != -1)
@@ -1938,7 +2006,7 @@ sigchld_early (int sig)
 }
 
 void
-setsigvhan (RETSIGTYPE (*handler) (int signo), int *sigv, int sigc)
+setsigvhan (void (*handler) (int signo), int *sigv, int sigc)
 {
   int i;
   struct sigaction act;
@@ -1982,7 +2050,7 @@ add_extra_sigv (int *sigv, int sigc)
 }
 
 void
-signal_setup (RETSIGTYPE (*sf) (int))
+signal_setup (void (*sf) (int))
 {
   setsigvhan (sf, default_sigv, ARRAY_SIZE (default_sigv));
   if (extra_sigc)
@@ -2561,6 +2629,10 @@ main (int argc, char **argv)
       components_trace (argv, depmap_col);
       exit (0);
 
+    case COM_LIST_SHUTDOWN_SEQUENCE:
+      components_list_shutdown_sequence ();
+      exit (0);
+
     default:
       pies_priv_setup (&pies_privs);
       if (pies_umask)
@@ -2650,10 +2722,12 @@ main (int argc, char **argv)
 	  component_config_commit ();
 	  if (SYSVINIT_ACTIVE)
 	    sysvinit_runlevel_setup (PIES_COMP_DEFAULT);
+	  if (children_op_toggle (PIES_CHLD_GC))
+	    progman_gc ();
 	  progman_create_sockets ();
 	  progman_start ();
 
-	  pies_schedule_children (PIES_CHLD_WAKEUP);
+	  //pies_schedule_children (PIES_CHLD_WAKEUP);
 	  action = ACTION_CONT;
 	  break;
 
@@ -2690,15 +2764,16 @@ main (int argc, char **argv)
 	}
       if (action == ACTION_CONT)
 	{
-	  if (children_op & PIES_CHLD_RESCHEDULE_ALARM)
-	    progman_recompute_alarm ();
-	  if (children_op & PIES_CHLD_GC)
+	  if (children_op_toggle (PIES_CHLD_GC))
 	    progman_gc ();
-	  if (children_op & PIES_CHLD_CLEANUP)
-	    progman_cleanup (0);
-	  if (children_op & PIES_CHLD_WAKEUP)
+	  if (children_op_toggle (PIES_CHLD_CLEANUP))
+	    progman_cleanup (CLEANUP_DEFAULT);
+	  /* The order is important here: progman_wake_sleeping
+	     can raise PIES_CHLD_RESCHEDULE_ALARM */
+	  if (children_op_toggle (PIES_CHLD_WAKEUP))
 	    progman_wake_sleeping (1);
-	  children_op = PIES_CHLD_NONE;
+	  if (children_op_toggle (PIES_CHLD_RESCHEDULE_ALARM))
+	    progman_recompute_alarm ();
 	}
     }
   while (SYSVINIT_ACTIVE || action == ACTION_CONT);
