@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2007-2020 Sergey Poznyakoff
+   Copyright (C) 2007-2025 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -100,6 +100,12 @@ prog_tag (struct prog const *prog)
    abort ();
  }
 
+int
+prog_sigterm (struct prog const *prog)
+{
+  return (prog && prog->type == TYPE_COMPONENT) ? prog->v.p.comp->sigterm : SIGTERM;
+}
+
 void
 link_prog (struct prog *prog, struct prog *ref)
 {
@@ -384,8 +390,8 @@ redirect_read (int fd, void *data)
     }
   else if (n == 0)
     {
-      logmsg (LOG_INFO, _("%s: EOF on %s"),
-	      tag, retr_stream_name[rb->stream]);
+      debug (3, (_("%s: EOF on %s"),
+		 tag, retr_stream_name[rb->stream]));
       rb->master->v.p.redir[rb->stream] = -1;
       close (fd);
       deregister_socket (fd);
@@ -1144,7 +1150,7 @@ prog_start (struct prog *prog)
 
       if (redir[RETR_OUT] != -1)
 	close (redir[RETR_OUT]);
-      if (redir[RETR_ERR])
+      if (redir[RETR_ERR] != -1)
 	close (redir[RETR_ERR]);
       prog->pid = pid;
       prog->v.p.status = status_running;
@@ -1596,20 +1602,20 @@ prog_stop_dependents (struct prog *prog)
   pies_depmap_pos_t pos;
   int warned = 0;
 
-  prog_stop_redirectors (prog);
   for (comp = component_depmap_first (depmap_row, prog->v.p.comp->arridx, &pos);
        comp;
        comp = component_depmap_next (pos))
     {
       struct prog *dp = comp->prog;
-      if (!dp) /* should not happen */
-	continue;
-      if (!warned && dp->pid)
+      if (dp)
 	{
-	  debug (1, ("stopping dependencies of %s", prog_tag (prog)));
-	  warned = 1;
+	  if (!warned && dp->pid)
+	    {
+	      debug (1, ("stopping dependencies of %s", prog_tag (prog)));
+	      warned = 1;
+	    }
+	  prog_stop (dp, prog_sigterm (dp));
 	}
-      prog_stop (dp, SIGTERM);
     }
   depmap_end (pos);
 }
@@ -1628,7 +1634,7 @@ prog_stop (struct prog *prog, int sig)
 	  pies_schedule_children (PIES_CHLD_RESCHEDULE_ALARM);
 	}
     }
-  debug (1, ("stopping %s (%lu)", prog_tag (prog), (unsigned long) prog->pid));
+  debug (1, ("sending signal %d to %s (%lu)", sig, prog_tag (prog), (unsigned long) prog->pid));
   if (prog->type == TYPE_COMPONENT && prog->v.p.comp->flags & CF_SIGGROUP)
     kill (-prog->pid, sig);
   else
@@ -1663,7 +1669,7 @@ print_status (const char *tag, pid_t pid, int status, int expect_term)
     {
       char const *coremsg = "";
 
-      if (expect_term && WTERMSIG (status) == SIGTERM)
+      if (expect_term && WTERMSIG (status) == expect_term)
 	prio = LOG_DEBUG;
 
 #ifdef WCOREDUMP
@@ -2004,14 +2010,33 @@ notify (const char *tag, int status, struct action *act)
 static int
 status_matches_p (struct action *act, unsigned status)
 {
-  int i;
-
-  if (act->nstat == 0)
-    return 1;
-  for (i = 0; i < act->nstat; i++)
-    if (act->status[i] == status)
-      return 1;
-  return 0;
+  int result;
+  
+  if (act->cond_list == NULL)
+    result = 1;
+  else
+    {
+      struct grecs_list_entry *ep;
+      
+      for (ep = act->cond_list->head; ep; ep = ep->next)
+	{
+	  struct status_cond *cp = ep->data;
+	  result = cp->status == status;
+	  if (cp->neg)
+	    {
+	      if ((status & STATUS_SIG_BIT) == (cp->status & STATUS_SIG_BIT))
+		{
+		  result = !result;
+		  if (!result)
+		    break;
+		}
+	    }
+	  else if (result)
+	    break;
+	}
+    }
+  
+  return result;
 }
 
 static void
@@ -2134,26 +2159,27 @@ react (struct prog *prog, int status, pid_t pid)
 }
 
 void
-progman_cleanup (int expect_term)
+progman_cleanup (int flags)
 {
   pid_t pid;
   int status;
+  int expect_term = (flags & CLEANUP_EXPECT_TERM) || progman_waiting_p ();
 
-  if (!expect_term)
-    expect_term = progman_waiting_p ();
   while ((pid = waitpid (-1, &status, WNOHANG)) > 0)
     {
       struct prog *prog = prog_lookup_by_pid (pid);
       if (!prog)
 	{
-	  print_status (_("unknown child"), pid, status, expect_term);
+	  print_status (_("unknown child"), pid, status,
+			expect_term ? SIGTERM : 0);
 	  continue;
 	}
       prog->pid = 0;
       switch (prog->type)
 	{
 	case TYPE_COMPONENT:
-	  print_status (prog_tag (prog), pid, status, expect_term);
+	  print_status (prog_tag (prog), pid, status,
+			expect_term ? prog_sigterm (prog) : 0);
 	  prog->stop = 0;
 	  if (prog->v.p.comp->mode == pies_comp_inetd)
 	    {
@@ -2214,8 +2240,9 @@ progman_cleanup (int expect_term)
 				   prog_tag (prog), pid, "");
 
 		  prog->v.p.status = status_stopped;
+		  prog_stop_redirectors (prog);
 		  prog_stop_dependents (prog);
-		  if (!expect_term)
+		  if (!expect_term && component_is_active (prog->v.p.comp))
 		    react (prog, status, pid);
 		}
 
@@ -2226,7 +2253,8 @@ progman_cleanup (int expect_term)
 	  break;
 
 	case TYPE_COMMAND:
-	  print_status (prog_tag (prog), pid, status, expect_term);
+	  print_status (prog_tag (prog), pid, status,
+			expect_term ? prog_sigterm (prog) : 0);
 	  destroy_prog (&prog);
 	  break;
 	}
@@ -2247,7 +2275,7 @@ progman_stop_component (struct prog **progptr)
 	{
 	case status_running:
 	  logmsg (LOG_INFO, _("stopping component %s"), prog_tag (prog));
-	  prog_stop (prog, SIGTERM);
+	  prog_stop (prog, prog_sigterm (prog));
 	  break;
 
 	case status_listener:
@@ -2332,6 +2360,28 @@ prog_to_stop (struct prog *prog)
   return prog;
 }
 
+static void
+timeout_init (int timeout, struct timespec *ts)
+{
+#ifdef HAVE_CLOCK_NANOSLEEP
+  clock_gettime (CLOCK_MONOTONIC, ts);
+  ts->tv_sec += timeout;
+#else
+  ts->tv_sec = timeout;
+  ts->tv_nsec = 0;
+#endif
+}
+
+static int
+timeout_expired (struct timespec *ts)
+{
+#ifdef HAVE_CLOCK_NANOSLEEP
+  return clock_nanosleep (CLOCK_MONOTONIC, TIMER_ABSTIME, ts, NULL) == 0;
+#else
+  return nanosleep (ts, ts) == 0;
+#endif
+}
+
 #define DIAG_CON \
   (SYSVINIT_ACTIVE ? (DIAG_TO_STDERR|DIAG_REOPEN_LOG) : diag_output)
 
@@ -2340,7 +2390,7 @@ prog_to_stop (struct prog *prog)
 void
 progman_gc (void)
 {
-  time_t start;
+  struct timespec ts;
   struct prog *prog, *next;
 
   /* Find first marked prog */
@@ -2348,10 +2398,8 @@ progman_gc (void)
   if (!prog)
     return;
 
-  /* First round: */
   /* Gracefully stop it and all marked programs that follow */
   diagmsg (DIAG_CON, LOG_INFO, "Sending processes the TERM signal");
-  start = time (NULL);
   do
     {
       next = prog->next;
@@ -2360,44 +2408,20 @@ progman_gc (void)
   while ((prog = prog_to_stop (next)));
 
   /* Wait for them to terminate */
-  while (1)
-    {
-      progman_cleanup (1);
-      prog = prog_to_stop (proghead);
-      if (!prog)
-	return;
-      if (time (NULL) - start < shutdown_timeout)
-	sleep (1);
-      else
-	break;
-    }
+  timeout_init (shutdown_timeout, &ts);
 
-  /* Second round: */
-  /* Kill the remaining programs with SIGKILL */
-  diagmsg (DIAG_CON, LOG_INFO, "Sending processes the KILL signal");
-  start = time (NULL);
   do
-    {
-      next = prog->next;
-      prog_stop (prog, SIGKILL);
-    }
-  while ((prog = prog_to_stop (next)));
-
-  /* Wait for them to terminate */
-  while (1)
     {
       progman_cleanup (1);
-      prog = prog_to_stop (proghead);
-      if (!prog)
+      if (!prog_to_stop (proghead))
 	return;
-      if (time (NULL) - start < shutdown_timeout)
-	sleep (1);
-      else
-	break;
     }
-
-  /* FIXME: Report remaining programs */
-
+  while (!timeout_expired (&ts));
+  
+  /* Kill the remaining programs with SIGKILL */
+  diagmsg (DIAG_CON, LOG_INFO, "Sending processes the KILL signal");
+  while ((prog = prog_to_stop (next)) != NULL)
+    prog_stop (prog, SIGKILL);
 }
 
 static int
@@ -2405,7 +2429,7 @@ start_shutdown (struct prog *prog, void *data)
 {
   if (IS_COMPONENT (prog) && prog->v.p.comp->mode == pies_comp_shutdown)
     {
-      int *p = data;
+      unsigned *p = data;
       if (!*p)
 	{
 	    diagmsg (DIAG_CON, LOG_INFO, "Starting shutdown components");
@@ -2418,41 +2442,110 @@ start_shutdown (struct prog *prog, void *data)
   return 0;
 }
 
-void
-program_shutdown (void)
+
+static void
+stop_progs_at_seqno (unsigned long seqno)
 {
-  time_t start = time (NULL);
-  int sd = 0;
+  struct prog *prog;
+  size_t count = 0;
+  struct timespec ts;
+  int warned = 0;
 
-  /* Activate shutdown components. */
-  progman_foreach (start_shutdown, &sd);
-  while (prog_to_stop (proghead))
+  /* Signal all programs that have this shutdown sequence number */
+  for (prog = proghead; prog; prog = prog->next)
     {
-      progman_cleanup (1);
-      if (time (NULL) - start < shutdown_timeout)
-	sleep (1);
-      else
+      if (IS_COMPONENT (prog) &&
+	  prog->v.p.status == status_running &&
+	  prog->v.p.comp->shutdown_seqno == seqno)
 	{
-	  progman_gc ();
-	  break;
+	  if (!warned)
+	    {
+	      debug (1, ("terminating processes at level %lu", seqno));
+	      warned = 1;
+	    }
+	  prog->stop = 1;
+	  prog_stop (prog, prog_sigterm (prog));
+	  count++;
 	}
     }
-}
 
-static int
-mark_for_stopping (struct prog *prog, void *data)
-{
-  if (IS_COMPONENT (prog))
-    prog->stop = 1;
-  return 0;
+  if (!count)
+    return;
+
+  debug (1, ("%zu processes signalled at level %lu", count, seqno));
+
+  timeout_init (shutdown_timeout, &ts);
+
+  /* Wait for them to terminate */
+  while (prog_to_stop (proghead))
+    {
+      if (timeout_expired (&ts))
+	{
+	  debug (1, ("killing %zu processes at level %lu", count, seqno));
+	  for (prog = proghead; prog; prog = prog->next)
+	    {
+	      if (IS_COMPONENT (prog) &&
+		  prog->v.p.comp->shutdown_seqno == seqno)
+		{
+		  prog_stop (prog, SIGKILL);
+		}
+	    }
+	  break;
+	}
+
+      progman_cleanup (CLEANUP_EXPECT_TERM);
+    }
 }
 
 void
 progman_stop (void)
 {
-  progman_foreach (mark_for_stopping, NULL);
-  progman_gc ();
-  program_shutdown ();
+  unsigned long *ordv;
+  size_t ordc;
+  size_t i;
+  int sd;
+
+  /*
+   * Stop all running components in correct order (dependents first).
+   */
+  ordc = components_shutdown_sequence_numbers (&ordv);
+
+  for (i = 0; i < ordc; i++)
+    {
+      stop_progs_at_seqno (ordv[i]);
+    }
+
+  /* Activate shutdown components. */
+  progman_foreach (start_shutdown, &sd);
+
+  /* Wait for them to finish. */
+  if (sd)
+    {
+      struct timespec ts;
+
+      timeout_init (shutdown_timeout, &ts);
+
+      /* Wait for them to terminate */
+      while (prog_to_stop (proghead))
+	{
+	  if (timeout_expired (&ts))
+	    {
+	      struct prog *prog;
+	      debug (1, ("killing shutdown processes"));
+	      for (prog = proghead; prog; prog = prog->next)
+		{
+		  if (IS_COMPONENT (prog) &&
+		      (prog->v.p.status == status_running
+		       || prog->v.p.status == status_stopping))
+		    {
+		      prog_stop (prog, SIGKILL);
+		    }
+		}
+	      break;
+	    }
+	  progman_cleanup (CLEANUP_EXPECT_TERM);
+	}
+    }
 }
 
 /* EOF */
