@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2007-2022 Sergey Poznyakoff
+   Copyright (C) 2007-2025 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -51,7 +51,8 @@ struct log_message_in
 /* Global variables */
 /* Fallback log file is used to log critical messages when syslog
    daemon is unavailable.  If NULL, stderr will be used. */
-char *pies_fallback_file = "/tmp/pies_logger.log";
+char *pies_fallback_file = FALLBACK_LOG;
+
 /* Log tag */
 char *pies_log_tag = "pies";
 /* Log facility */
@@ -67,8 +68,14 @@ size_t pies_log_max_queue = PIES_LOG_MAX_QUEUE;
    The user can modify this value using the pies_syslog_set_dev call.
 */
 static char *log_dev;
+
+enum
+  {
+    LOGFD_NOT_OPEN = -1,
+    LOGFD_DISABLED = -2
+  };
 /* Log socket descriptor. */
-static int log_fd = -1;
+static int log_fd = LOGFD_NOT_OPEN;
 
 /* Socket address */
 static union {
@@ -119,16 +126,17 @@ reopen_logger (void)
 {
   int fd;
   int flags;
+  char *dev = log_dev ? log_dev : PIES_LOG_DEV;
   
   if (log_salen == 0)
     {
-      char *dev = log_dev ? log_dev : PIES_LOG_DEV;
       if (dev[0] == '/')
 	{
 	  size_t len = strlen (dev);
 	  if (len >= sizeof log_sa.s_un.sun_path)
 	    {
 	      fallback_log ("%s: UNIX socket name too long", dev);
+	      log_fd = LOGFD_DISABLED;
 	      return -1;
 	    }
 	  strcpy (log_sa.s_un.sun_path, dev);
@@ -164,6 +172,7 @@ reopen_logger (void)
 	  if (rc)
 	    {
 	      fallback_log ("%s: invalid socket address", dev);
+	      log_fd = LOGFD_DISABLED;
 	      return -1;
 	    }
 	  
@@ -178,7 +187,19 @@ reopen_logger (void)
 
   if (fd == -1)
     {
-      fallback_log ("socket: %s", strerror (errno));
+      fallback_log ("can't create socket (family %d): %s",
+		    log_family, strerror (errno));
+      switch (errno)
+	{
+	case EAGAIN:
+	case EINTR:
+	case EMFILE:
+	case ENFILE:
+	  break;
+
+	default:
+	  log_fd = LOGFD_DISABLED;
+	}
       return -1;
     }
 
@@ -188,19 +209,30 @@ reopen_logger (void)
       fcntl (fd, F_SETFD, flags | FD_CLOEXEC) == -1)
     {
       close (fd);
+      log_fd = LOGFD_DISABLED;
       return -1;
     }
     
-    if (connect(fd, (struct sockaddr*)&log_sa, log_salen))
-      {
-	fallback_log("socket: %s", strerror(errno));
-	close(fd);
-	return -1;
+  if (connect (fd, (struct sockaddr*)&log_sa, log_salen))
+    {
+      fallback_log ("can't connect to log socket %s: %s", dev,
+		    strerror (errno));
+      close (fd);
+      switch (errno) {
+      case EAGAIN:
+      case EINTR:
+      case ECONNREFUSED:
+	break;
+	
+      default:
+	log_fd = LOGFD_DISABLED;
       }
+      return -1;
+    }
 
-    log_fd = fd;
-    
-    return 0;
+  log_fd = fd;
+  
+  return 0;
 }
 
 static struct log_message_in *
@@ -390,9 +422,17 @@ pies_syslog_flush (void)
   
   while ((msg = log_message_dequeue ()) != NULL)
     {
-      if (log_fd == -1)
-	reopen_logger ();
+      switch (log_fd)
+	{
+	case LOGFD_NOT_OPEN:
+	  reopen_logger ();
+	  break;
 
+	case LOGFD_DISABLED:
+	  fallback_log ("%*.*s", msg->len, msg->len, msg->text);
+	  continue;
+	}
+      
       rc = send (log_fd, msg->text, msg->len, MSG_NOSIGNAL);
       if (rc != -1)
 	{
@@ -415,7 +455,7 @@ pies_syslog_flush (void)
       if (errno == EPIPE)
 	{
 	  close (log_fd);
-	  log_fd = -1;
+	  log_fd = LOGFD_NOT_OPEN;
 	  continue;
 	}
 
@@ -442,7 +482,6 @@ pies_syslog_flush (void)
 	}
 		
       /* Else ? */
-      break;
     }
 }
 
@@ -483,11 +522,9 @@ pies_syslog_open (void)
 void
 pies_syslog_close (void)
 {
-  if (log_fd != -1)
-    {
-      close (log_fd);
-      log_fd = -1;
-    }
+  if (log_fd >= 0)
+    close (log_fd);
+  log_fd = LOGFD_NOT_OPEN;
 }
 
 void
